import typing
from abc import ABCMeta, abstractmethod as abstract_method
from collections import defaultdict as default_dict

from data_providers import BaseDataProvider, MutableDataProviderMixin
from data_providers.data_structures import ParameterisedDefaultDict
from data_providers.exceptions import with_context

__all__ = [
    'BaseDataProviderDelegate',
    'MutableDataProviderDelegateMixin',
]


class BaseDataProviderDelegate(metaclass=ABCMeta):
    """
    A data-provider-like object that delegates its responsibilities to other
    data providers.

    The delegate stores a collection of data providers internally, and it
    routes each incoming value to the appropriate data provider.
    """

    def __init__(self):
        super(BaseDataProviderDelegate, self).__init__()

        self._delegates: typing.Dict[
            typing.Hashable,
            BaseDataProvider,
        ] = ParameterisedDefaultDict(self.create_data_provider)

    def __getitem__(self, value: typing.Any) -> typing.Any:
        """
        Returns the data for the specified value, fetching it from the backend
        if necessary.
        """
        return self.get_data_provider(value)[value]

    @abstract_method
    def gen_delegate_key(self, value: typing.Any) -> typing.Hashable:
        """
        Generates a delegate key for a registered value.

        :param value:
            The value passed to `__getitem__`.
        """
        raise NotImplementedError(
            'Not implemented in {cls}.'.format(cls=type(self).__name__),
        )

    @abstract_method
    def create_data_provider(self,
            delegate_key: typing.Hashable,
    ) -> BaseDataProvider:
        """
        Creates the appropriate data provider for the specified delegate key.

        If ``delegate_key`` is invalid, this method must raise an exception.

        .. note::

           The resulting data provider instance is cached locally; this method
           is only invoked once per delegate key.
        """
        raise NotImplementedError(
            'Not implemented in {cls}.'.format(cls=type(self).__name__),
        )

    def register(self, values: typing.Iterable[typing.Any]) -> None:
        """
        Registers a set of values with the data provider.
        """
        grouped_values = self.group_by_delegate_key(values)

        for delegate_key, delegate_values in grouped_values.items():
            self._delegates[delegate_key].register(delegate_values)

    def get_data_provider(self, value: typing.Any) -> BaseDataProvider:
        """
        Returns the correct data provider to use for the specified value.
        """
        return self._delegates[self.gen_delegate_key(value)]

    def group_by_delegate_key(self,
            values: typing.Iterable[typing.Any],
    ) -> typing.Mapping[typing.Hashable, typing.Iterable[typing.Any]]:
        """
        Groups a collection of values by delegate key.

        This method may be overridden in subclasses if desired.
        """
        grouped = default_dict(list)

        for value in values:
            grouped[self.gen_delegate_key(value)].append(value)

        return grouped


class MutableDataProviderDelegateMixin(
    BaseDataProviderDelegate,
    metaclass=ABCMeta,
):
    """
    A mixin for data provider delegates that allow external code to modify the
    contents of the data providers' caches.
    """

    def __setitem__(self, value: typing.Any, cache_value: typing.Any):
        """
        Sets the cached value for the specified object.
        """
        delegate_key = self.gen_delegate_key(value)
        data_provider = self._delegates[delegate_key]

        if isinstance(data_provider, MutableDataProviderMixin):
            data_provider[value] = cache_value
        else:
            raise with_context(
                TypeError(
                    f'Cannot modify the contents of '
                    f'{type(data_provider).__name__}.'
                ),

                context={
                    'data_provider': data_provider,
                    'delegate_key': delegate_key,
                    'value': value,
                },
            )



from unittest import TestCase

from data_providers import MutableDataProviderMixin
from data_providers.delegate import BaseDataProviderDelegate, \
    MutableDataProviderDelegateMixin
from data_providers.testing import MockDataProvider


class ProfessionalDataProvider(MockDataProvider):
    """
    Template for :py:class:`TestDelegate`.
    """

    def gen_load_key(self, value):
        return value['lastName']

    def gen_empty_result(self):
        return {}


class TestDelegate(BaseDataProviderDelegate):
    """
    Data provider delegate used by :py:class:`DataProviderDelegateTestCase`.
    """

    def gen_delegate_key(self, value):
        return value['profession']

    def create_data_provider(self, profession):
        if profession == 'historian':
            return ProfessionalDataProvider({
                'Jones': {'firstName': 'Henry'},
                'Brody': {'firstName': 'Marcus'},
            })

        elif profession == 'nazi_stooge':
            return ProfessionalDataProvider({
                'Belloq': {'firstName': 'René'},
                'Donovan': {'firstName': 'Walter'},
            })

        raise ValueError(profession)


class MutableProfessionalDataProvider(
    # Mixin listed second so that it doesn't override MRO for subclass (it's
    # closer to the base class py:class:`BaseDataProvider` than
    # :py:class:`ProfessionalDataProvider` is).
    ProfessionalDataProvider,
    MutableDataProviderMixin,
):
    """
    Mutable version of :py:class:`ProfessionalDataProvider`.
    """
    pass


class TestMutableDelegate(
    # Mixin listed second so that it doesn't override MRO for subclass (it's
    # closer to the base class py:class:`BaseDataProvider` than
    # :py:class:`TestDelegate` is).
    TestDelegate,
    MutableDataProviderDelegateMixin,
):
    """
    Mutable version of :py:class:`TestDelegate`.
    """

    def create_data_provider(self, profession):
        if profession == 'historian':
            return MutableProfessionalDataProvider({
                'Jones': {'firstName': 'Henry'},
                'Brody': {'firstName': 'Marcus'},
            })

        elif profession == 'nazi_stooge':
            # Note that this data provider is not mutable.
            return ProfessionalDataProvider({
                'Belloq': {'firstName': 'René'},
                'Donovan': {'firstName': 'Walter'},
            })

        raise ValueError(profession)


class DataProviderDelegateTestCase(TestCase):
    """
    Tests various aspects of data provider delegate implementation (effectively
    acts as a test case for :py:class:`BaseDataProviderDelegate`).
    """

    def test_happy_path(self):
        """
        Typical usage of data provider delegate.
        """
        users = [
            {'lastName': 'Jones', 'profession': 'historian'},
            {'lastName': 'Brody', 'profession': 'historian'},
            {'lastName': 'Belloq', 'profession': 'nazi_stooge'},
            {'lastName': 'Donovan', 'profession': 'nazi_stooge'},
        ]

        # Delegates look and behave exactly like data providers.
        data_provider = TestDelegate()
        data_provider.register(users)

        for u in users:
            u.update(data_provider[u])

        self.assertDictEqual(
            users[0],

            {
                'firstName': 'Henry',
                'lastName': 'Jones',
                'profession': 'historian',
            },
        )

        self.assertDictEqual(
            users[1],

            {
                'firstName': 'Marcus',
                'lastName': 'Brody',
                'profession': 'historian',
            },
        )

        self.assertDictEqual(
            users[2],

            {
                'firstName': 'René',
                'lastName': 'Belloq',
                'profession': 'nazi_stooge',
            },
        )

        self.assertDictEqual(
            users[3],

            {
                'firstName': 'Walter',
                'lastName': 'Donovan',
                'profession': 'nazi_stooge',
            },
        )

    def test_invalid_delegate_key(self):
        """
        A value cannot be matched with an appropriate data provider.
        """
        data_provider = TestDelegate()

        with self.assertRaises(ValueError):
            data_provider.register([
                # This value can't be registered because
                # :py:class:`TestDelegate` doesn't know which data provider to
                # use for the 'adventurer' profession.
                {'lastName': 'Ravenwood', 'profession': 'adventurer'},
            ])


class MutableDataProviderDelegateTestCase(TestCase):
    """
    Tests various aspects of mutable data provider delegate implementation
    (effectively acts as a test case for
    :py:class:`MutableDataProviderDelegateMixin`).
    """

    def test_add_arbitrary_values(self):
        """
        Adding arbitrary values to data provider caches.
        """
        users = [
            {'lastName': 'Brody', 'profession': 'historian'},
            {'lastName': 'Belloq', 'profession': 'nazi_stooge'},
        ]

        data_provider = TestMutableDelegate()
        data_provider.register(users)

        wild_card = {'lastName': 'Jones', 'profession': 'historian'}

        # If we try to get data for an unregistered value, we get an exception
        # as normal.
        with self.assertRaises(ValueError):
            # noinspection PyStatementEffect
            data_provider[wild_card]

        # However, if we specify a cache value explicitly, then we can retrieve
        # it - even if we haven't loaded anything yet.
        data_provider[wild_card] = {'firstName': 'Henry'}

        self.assertDictEqual(
            data_provider[wild_card],
            {'firstName': 'Henry'},
        )

        # We can override this value any time we want to.
        data_provider[wild_card] = {'firstName': 'Indiana'}

        self.assertDictEqual(
            data_provider[wild_card],
            {'firstName': 'Indiana'},
        )

        # Other values continue to work as expected.
        self.assertDictEqual(
            data_provider[users[0]],
            {'firstName': 'Marcus'},
        )

        with self.assertRaises(ValueError):
            # noinspection PyStatementEffect
            data_provider[{'lastName': 'Donovan', 'profession': 'nazi_stooge'}]

    def test_override_backend_values(self):
        """
        Overriding data returned by the backends.
        """
        users = [
            {'lastName': 'Jones', 'profession': 'historian'},
            {'lastName': 'Brody', 'profession': 'historian'},
        ]

        data_provider = TestMutableDelegate()
        data_provider.register(users)

        # I've got a lot of fond memories of that dog.
        data_provider[users[0]] = {'firstName': 'Indiana'}

        self.assertDictEqual(
            data_provider[users[0]],
            {'firstName': 'Indiana'},
        )

        # Other values continue to work as expected.
        self.assertDictEqual(
            data_provider[users[1]],
            {'firstName': 'Marcus'},
        )

        with self.assertRaises(ValueError):
            # noinspection PyStatementEffect
            data_provider[{'lastName': 'Donovan', 'profession': 'nazi_stooge'}]

    def test_immutable_data_provider(self):
        """
        Attempting to override a value for an immutable data provider.
        """
        data_provider = TestMutableDelegate()

        user = {'lastName': 'Belloq', 'profession': 'nazi_stooge'}

        # Note that :py:class:`TestMutableDelegate` uses an immutable data
        # provider for the "nazi_stooge" profession.
        with self.assertRaises(TypeError):
            data_provider[user] = {'firstName': 'René'}

    def test_invalid_delegate_key(self):
        """
        Attempting to override a value that cannot be matched with an
        appropriate data provider.
        """
        data_provider = TestMutableDelegate()

        # This value won't work because :py:class:`MutableTestDelegate` doesn't
        # know which data provider to use for the 'adventurer' profession.
        user = {'lastName': 'Ravenwood', 'profession': 'adventurer'}

        with self.assertRaises(ValueError):
            data_provider[user] = {'firstName': 'Marion'}



import typing
from unittest import TestCase, mock

from data_providers.data_structures import ParameterisedDefaultDict


class ParameterizedDefaultDictTestCase(TestCase):
    @staticmethod
    def test_no_params():
        """
        The factory only receives the missing key when it is invoked.
        """
        factory = mock.Mock(return_value=[])

        d: typing.Dict[
            typing.Text,
            typing.List[typing.Text],
        ] = ParameterisedDefaultDict(factory)

        d['foo'].append('bar')
        d['foo'].append('baz')
        d['foobie'].append('bletch')

        factory.assert_has_calls([
            mock.call('foo'),
            mock.call('foobie'),
        ])

    @staticmethod
    def test_with_params():
        """
        Specifying additional params to pass to the factory.
        """
        factory = mock.Mock(return_value=[])

        d: typing.Dict[
            typing.Text,
            typing.List[typing.Text],
        ] = ParameterisedDefaultDict(
            factory,

            # Additional positional arguments.
            'alpha',
            'bravo',

            # Additional keyword arguments.
            charlie='delta',
            echo='foxtrot',
        )

        d['foo'].append('bar')
        d['foo'].append('baz')
        d['foobie'].append('bletch')

        factory.assert_has_calls([
            mock.call('foo', 'alpha', 'bravo', charlie='delta',
                echo='foxtrot'),
            mock.call('foobie', 'alpha', 'bravo', charlie='delta',
                echo='foxtrot'),
        ])



import typing
from collections import defaultdict as default_dict

T = typing.TypeVar('T')


class ParameterisedDefaultDict(default_dict):
    """
    A :py:class:`default_dict` that passes parameters to its
    :py:meth:`__missing__` method.
    """

    def __init__(self,
            default_factory: typing.Optional[typing.Callable[..., T]] = None,
            *args: typing.Any,
            **kwargs: typing.Any,
    ) -> None:
        """
        :param default_factory:
            Factory method/class that will be used to create new items
            automatically.

            Must accept at least one argument (the key of the new value).

        :param args:
            Additional positional arguments that will be passed to the factory
            after the key value.

        :param kwargs:
            Additional keyword arguments that will be passed to the factory.
        """
        super(ParameterisedDefaultDict, self).__init__(default_factory)

        self._factory_args = args
        self._factory_kwargs = kwargs

    def __missing__(self, key: typing.Hashable) -> T:
        """
        Invoked when accessing a key that does not yet exist in the dict.
        """
        if not self.default_factory:
            raise KeyError(key)

        # noinspection PyArgumentList
        self[key] = value = self.default_factory(
            key,
            *self._factory_args,
            **self._factory_kwargs
        )

        return value



from unittest import TestCase

from data_providers.testing import MockDataProvider


class MockDataProviderTestCase(TestCase):
    def test_happy_path(self):
        """
        Typical usage of :py:class:`MockDataProvider`.
        """
        users = {
            'alpha': {},
            'bravo': {},
            'charlie': {},
            'delta': {},
        }

        data_provider = MockDataProvider({
            # Note that the backend will not return any data for ``delta``.
            'alpha': {'firstName': 'Henry'},
            'bravo': {'firstName': 'Henry'},
            'charlie': {'firstName': 'Marcus'},
        })

        # You can't get any data until you've registered the corresponding
        # values.
        with self.assertRaises(ValueError):
            # noinspection PyStatementEffect
            data_provider['alpha']

        data_provider.register(users.keys())

        for username, user_data in users.items():
            user_data.update(data_provider[username])

        self.assertDictEqual(users['alpha'], {'firstName': 'Henry'})
        self.assertDictEqual(users['bravo'], {'firstName': 'Henry'})
        self.assertDictEqual(users['charlie'], {'firstName': 'Marcus'})

        # The backend didn't return any data for ``delta``.
        self.assertDictEqual(users['delta'], {})



from data_providers import BaseDataProvider

__all__ = [
    'MockDataProvider',
]


class MockDataProvider(BaseDataProvider):
    """
    A data provider that "loads" data from a dict stored in memory.

    This class is generally only used during unit tests.
    """

    def __init__(self, data):
        """
        :param data:
            Data used to seed the "backend".

            Usually, this is a list of dicts (simulates the result of a SQL
            query, data frame from an R function, etc.).
        """
        super(MockDataProvider, self).__init__()

        self._data = data

    def fetch_from_backend(self, load_keys):
        # If a load key does not appear in :py:attr:`_data`, do not return any
        # data for it.  This allows us to simulate that specific case during
        # unit tests.
        return (
            (lk, self._data[lk])
            for lk in load_keys
            if lk in self._data
        )

    def gen_empty_result(self):
        return {}



import typing

TException = typing.TypeVar('TException', bound=Exception)


def with_context(
        exc: TException,
        context: typing.Mapping[typing.Text, typing.Any],
) -> TException:
    """
    Attaches a context dict to an exception, so that it can be captured by
    loggers.

    This lets you keep the exception message fairly short/generic, while making
    useful troubleshooting information accessible to debuggers and loggers.

    Example:

    .. code-block:: python

       raise with_context(
         ValueError('Failed validation.'),

         context = {
            'errors': filter_runner.get_context(True),
         },
       )
    """
    if not hasattr(exc, 'context'):
        exc.context = {}

    if not isinstance(exc.context, typing.MutableMapping):
        exc.context = {'_context': exc.context}

    exc.context.update(context)

    return exc



from itertools import chain
from unittest import TestCase

from data_providers.aggregator import BaseDataProviderAggregator
from data_providers.testing import MockDataProvider


class MovieLineDataProvider(MockDataProvider):
    """
    Data provider that will be used to test
    :py:class:`BaseDataProviderAggregator`.
    """

    def gen_load_key(self, value):
        return value['lastName']

    def gen_empty_result(self):
        return []


class MovieLineAggregator(BaseDataProviderAggregator):
    """
    Trivial data provider aggregator implementation that assembles movie lines
    spoken by celebrities from a variety of sources.

    Used to test :py:class:`BaseDataProviderAggregator`.
    """

    def aggregate_data(self, value, data):
        return list(chain(*data.values()))

    def create_data_providers(self):
        # Simulate a scenario where we are loading movie lines from different
        # sources depending on which movie they come from.
        return {
            # Maybe we have all the lines from Brazil stored in a local
            # database; I'm sure somebody does, anyway.
            'brazil':
                MovieLineDataProvider({
                    'Pryce': [
                        'Triplets? My how time flies.',
                        "It's my lunch hour. Besides, it's not my department.",
                    ],

                    'De Niro': [
                        "Listen, kid, we're all in it together.",
                        'My good friends call me Harry.',
                    ],
                }),

            # Perhaps we hit a 3rd-party web service to download lines from
            # Taxi Driver.
            'taxiDriver':
                MovieLineDataProvider({
                    'De Niro': [
                        "You're only as healthy as you feel.",
                    ],

                    'Keitel': [
                        "You're a funny guy - but looks aren't everything.",
                    ],
                }),
        }

    def gen_routing_keys(self, value):
        if value['lastName'] in {'De Niro', 'Pryce'}:
            yield 'brazil'

        if value['lastName'] in {'De Niro', 'Keitel'}:
            yield 'taxiDriver'


class DataProviderAggregatorTestCase(TestCase):
    def test_happy_path(self):
        """
        Typical usage of a data provider aggregator.
        """
        jonathan = {'firstName': 'Jonathan', 'lastName': 'Pryce'}
        harvey = {'firsName': 'Harvey', 'lastName': 'Keitel'}
        robert = {'firstName': 'Robert', 'lastName': 'De Niro'}

        aggregator = MovieLineAggregator()
        aggregator.register([jonathan, harvey, robert])

        self.assertEqual(
            aggregator[jonathan],

            [
                'Triplets? My how time flies.',
                "It's my lunch hour. Besides, it's not my department.",
            ],
        )

        self.assertEqual(
            aggregator[harvey],

            [
                "You're a funny guy - but looks aren't everything.",
            ],
        )

        self.assertEqual(
            aggregator[robert],

            # Results from multiple data providers are aggregated.
            [
                "Listen, kid, we're all in it together.",
                'My good friends call me Harry.',
                "You're only as healthy as you feel.",
            ],
        )



import typing
from abc import ABCMeta, abstractmethod as abstract_method
from collections import OrderedDict

from data_providers import BaseDataProvider

__all__ = [
    'BaseDataProviderAggregator',
]


class BaseDataProviderAggregator(metaclass=ABCMeta):
    """
    A data-provider-like object that aggregates values from several data
    providers.
    """

    def __init__(self):
        super(BaseDataProviderAggregator, self).__init__()

        self._data_providers = None

    def __getitem__(self, value):
        """
        Returns the data for the specified value.

        :param value:
            The lookup key.

        :raise:
            - :py:class:`ValueError` if the value wasn't registered first.
        """
        return self.aggregate_data(
            value=value,

            data={
                key: self.data_providers[key][value]
                for key in self.gen_routing_keys(value)
            },
        )

    @property
    def data_providers(self) -> typing.Mapping[
        typing.Hashable,
        BaseDataProvider,
    ]:
        """
        Lazy-loads the data providers for this aggregator.
        """
        if self._data_providers is None:
            self._data_providers = self.create_data_providers()
        return self._data_providers

    @abstract_method
    def aggregate_data(self,
            value: typing.Any,
            data: typing.Mapping[typing.Hashable, typing.Iterable[typing.Any]],
    ) -> typing.Any:
        """
        Aggregates the data returned by each of the data providers for the
        specified value.

        :param value:
            The value provided to :py:meth:`__getitem__`.

        :param data:
            A mapping containing the data returned by each data provider
            (corresponding to :py:meth:`create_data_providers`).

            Note that this value might only contain a subset of the data
            providers, depending on :py:meth:`gen_routing_keys`.
        """
        raise NotImplementedError(
            'Not implemented in {cls}.'.format(cls=type(self).__name__),
        )

    @abstract_method
    def create_data_providers(self) -> typing.Mapping[
        typing.Hashable,
        BaseDataProvider,
    ]:
        """
        Creates the mapping of data providers used by this aggregator.

        :return:
            Must return a mapping, so that each data provider is mapped to a
            unique key.
        """
        raise NotImplementedError(
            'Not implemented in {cls}.'.format(cls=type(self).__name__),
        )

    # noinspection PyUnusedLocal
    def gen_routing_keys(self,
            value: typing.Any,
    ) -> typing.Iterable[typing.Hashable]:
        """
        Returns the keys of the data providers that should handle the specified
        value.

        :return:
            Values correspond to keys in the mapping returned by
            :py:meth:`create_data_providers`.
        """
        return self.data_providers.keys()

    def group_by_routing_keys(self,
            values: typing.Iterable[typing.Any],
    ) -> typing.Mapping[typing.Hashable, typing.Iterable[typing.Any]]:
        """
        Groups a set of values by their routing keys.
        """
        # Using an OrderedDict, just in case order is significant.
        groups = OrderedDict()

        for v in values:
            for key in self.gen_routing_keys(v):
                groups.setdefault(key, []).append(v)

        return groups

    def register(self, values: typing.Iterable[typing.Any]) -> None:
        """
        Registers a set of values so that the data providers can plan out the
        bulk queries they need to execute against the backend.
        """
        for key, group in self.group_by_routing_keys(values).items():
            self.data_providers[key].register(group)



import typing
from abc import ABCMeta, abstractmethod as abstract_method
from collections import defaultdict as default_dict

from data_providers.exceptions import with_context

__all__ = [
    'BaseDataProvider',
    'MutableDataProviderMixin',
]


class BaseDataProvider(metaclass=ABCMeta):
    """
    Implementation of the flywheel pattern.  Bulk-loads data from a backend
    and parcels them out on an as-needed basis.
    """

    def __init__(self):
        super(BaseDataProvider, self).__init__()

        self._cache = self._create_cache()
        """
        Caches data loaded from the backend so that it can be accessed easily.
        """

        self._pending_load_keys: typing.Set[typing.Hashable] = set()
        """
        Stores lookup keys for values that haven't been loaded yet.

        Note: This value should be a set; if you find yourself wanting to use a
        dict here, :py:class:`BaseDataProviderDelegate` is probably a better
        fit for your use case.
        """

        self._pending_cache_keys: typing.Dict[
            typing.Hashable,
            typing.Set[typing.Hashable]
        ] = default_dict(set)
        """
        Stores cache keys for values that haven't been loaded yet.

        This is used to gracefully handle cases where the backend doesn't
        return data for one or more load keys.
        """

    def __getitem__(self, value: typing.Any) -> typing.Any:
        """
        Returns data for the specified value.

        :param value:
            The lookup key.

        :raise:
            - :py:class:`ValueError` if the value wasn't registered first.
        """
        # Determine the key we will use to look up the value in the cache.
        cache_key = self.gen_cache_key(value)
        if cache_key is None:
            return self.gen_empty_result()

        try:
            cached = self._get_from_cache(cache_key)
        except KeyError:
            # The value has not been cached yet.
            # Determine which data we need to load from the backend.
            load_key = self.gen_load_key(value)
            if load_key is None:
                cached = self.gen_empty_result()
            else:
                # If the key hasn't been registered, we cannot continue.
                # If we were to allow loading without registering, we would
                # undermine the whole point of using a flywheel in the first
                # place!
                if load_key not in self._pending_load_keys:
                    raise with_context(
                        ValueError(
                            f'Attempting to get data for '
                            'unregistered value {value!r}.',
                        ),

                        context={
                            'cacheKey': cache_key,
                            'loadKey': load_key,
                            'value': value,
                        },
                    )

                # Load data from the backend and populate/extend the cache.
                self._load_data(load_key)

                # Try loading from cache again, this time without a safety net.
                cached = self._get_from_cache(cache_key)

        return cached

    @abstract_method
    def fetch_from_backend(self, load_keys: typing.Set[typing.Hashable]) -> \
            typing.Union[
                typing.Mapping[typing.Hashable, typing.Any],
                typing.Generator[
                    typing.Tuple[typing.Hashable, typing.Any], None, None],
            ]:
        """
        Fetches data from the backend for the specified load keys.

        :return:
            A mapping, or a generator that yields (key, value) pairs.

            The keys in the resulting mapping should correspond to the data
            provider's cache keys (from :py:meth:`gen_cache_key`).
        """
        raise NotImplementedError(
            'Not implemented in {cls}.'.format(cls=type(self).__name__),
        )

    def gen_load_key(self,
            value: typing.Any,
    ) -> typing.Optional[typing.Hashable]:
        """
        Extracts the value that will be used by the backend to load
        data for the specified value.

        If this method returns ``None``, then the data provider will
        never attempt to return data for that value; if that value is
        passed to :py:meth:`__getitem__`, the data provider will return
        an empty result instead (see :py:meth:`gen_empty_result`).

        :param value:
            The value passed to :py:meth:`__getitem__`.
        """
        return value

    def gen_cache_key(self,
            value: typing.Any,
    ) -> typing.Optional[typing.Hashable]:
        """
        Generates a lookup key for the data provider's cached data.

        If this method returns ``None``, then the data provider will never
        attempt to return data for that value; if that value is passed to
        :py:meth:`__getitem__`, the data provider will return an empty result
        instead (see :py:meth:`gen_empty_result`).

        :param value:
            The key value passed to `__getitem__`.
        """
        return self.gen_load_key(value)

    def gen_empty_result(self) -> typing.Optional[typing.Any]:
        """
        Generates an empty result that is returned if the backend cannot (or
        will not) return data for a particular key.

        You may override this method in a subclass if desired.  This method may
        return any value, or even raise an exception, depending on the use
        case.
        """
        return None

    def register(self, values: typing.Iterable[typing.Any]) -> None:
        """
        Registers a set of values so that the data provider can plan out the
        bulk queries it needs to execute against the backend.
        """
        for v in values:
            self._add_to_key_registry(v)

    def warm_cache(self):
        """
        Warms the cache, fetching data for any keys that haven't been loaded
        yet.

        Note: the data provider will automatically load data from the backend
        as needed.  In general, you shouldn't have to invoke this method,
        though it can be useful in particular during debugging.
        """
        if self._pending_load_keys:
            self._load_batch(self._pending_load_keys)

    # noinspection PyMethodMayBeStatic
    def _create_cache(self) -> typing.MutableMapping[
        typing.Hashable,
        typing.Any,
    ]:
        """
        Initializes the cache that the data provider will use.

        You may override this method in a subclass if desired; the only
        restriction is that it must return some kind of mutable mapping.
        """
        return {}

    def _add_to_cache(self,
            cache_key: typing.Hashable,
            cache_value: typing.Any,
    ) -> None:
        """
        Adds a value to the cache.

        If there is an existing value with the same key, it is overwritten.

        You may override this method in a subclass if desired.
        """
        self._cache[cache_key] = cache_value

    def _get_from_cache(self, cache_key: typing.Hashable) -> typing.Any:
        """
        Attempts to retrieve the data from the cache for the specified key.

        You may override this method in a subclass if desired.

        :raise:
          - :py:class:`KeyError` in the event of a cache miss.
        """
        return self._cache[cache_key]

    def _add_to_key_registry(self, value: typing.Any) -> None:
        """
        Adds the specified value to the registry of unloaded keys.

        :param value:
            The value that was passed to :py:meth:`register`.
        """
        # If we can't load data for a value, then there's no point in
        # registering it.
        load_key = self.gen_load_key(value)
        if load_key is not None:
            # If we can't cache data for a value, then there would be no point
            # in loading data for it.
            # Also, we don't need to register a value if we already have cached
            # data for it.
            cache_key = self.gen_cache_key(value)
            if (cache_key is not None) and (cache_key not in self._cache):
                # Queue the load key for the next time we call
                # :py:meth:`_load_data`.
                self._pending_load_keys.add(load_key)

                # Queue the corresponding cache key so that we can ensure that
                # *some* kind of value is cached, even if the backend doesn't
                # return any data for it.
                self._pending_cache_keys[load_key].add(cache_key)

    # noinspection PyUnusedLocal
    def _get_keys_to_load(self, hint: typing.Hashable) -> typing.Set[
        typing.Hashable]:
        """
        Used by :py:meth:`_load_data` to decide which values to send to the
        backend.

        By default, this method returns all pending load keys, but you may
        override this method in a subclass if you want to use a more selective
        strategy.

        :param hint:
            The requested load key (based on the value passed to
            :py:meth:`__getitem__`).
        """
        return self._pending_load_keys

    def _load_data(self, hint: typing.Hashable) -> None:
        """
        Bulk-loads data from the backend in the event of a cache miss.

        :param hint:
            The requested load key (based on the value passed to
            :py:meth:`__getitem__`).
        """
        batched_load_keys = self._get_keys_to_load(hint)
        if batched_load_keys:
            self._load_batch(batched_load_keys)

    def _load_batch(self, load_keys: typing.Set[typing.Hashable]) -> None:
        """
        Bulk-loads data from the backend, only for the specified keys.
        """
        loaded = self.fetch_from_backend(load_keys)

        # :py:meth:`fetch_from_backend` is expected to return a mapping, or at
        # least a value that quacks like one.
        if isinstance(loaded, typing.Generator):
            iterator = loaded
        else:
            if not isinstance(loaded, typing.Mapping):
                loaded = dict(loaded)

            iterator = loaded.items()

        for cache_key, cache_value in iterator:
            # The backend can overwrite any existing cached values.
            # You may override :py:meth:`_add_to_cache` in a subclass if this
            # is not desirable behavior.
            self._add_to_cache(cache_key, cache_value)

        # Fill in any missing cache keys that weren't included in the result
        # from the backend.
        for lk in load_keys:
            for pck in self._pending_cache_keys[lk]:
                if pck not in self._cache:
                    self._add_to_cache(pck, self.gen_empty_result())
            self._pending_cache_keys.pop(lk, None)

        self._pending_load_keys -= load_keys


class MutableDataProviderMixin(BaseDataProvider, metaclass=ABCMeta):
    """
    A mixin for data provider classes that allow external code to modify the
    contents of the cache.

    Any values set this way will override values returned by the backend.
    """

    def __init__(self):
        super(MutableDataProviderMixin, self).__init__()

        self._override_cache_keys: typing.Set[typing.Hashable] = set()

    def __setitem__(self, value: typing.Any, cache_value: typing.Any) -> None:
        """
        Sets the cached value for the specified object.

        :param value:
            Same as :py:meth:`__getitem__`.

        :param cache_value:
            The value that should be put in the cache.
        """
        load_key = self.gen_load_key(value)
        if load_key is not None:

            cache_key = self.gen_cache_key(value)
            if cache_key is not None:

                # Since we are setting the cached value explicitly, we do not
                # need to backfill it when :py:meth:`_load_data` runs.
                pending_cache_keys = self._pending_cache_keys[load_key]
                pending_cache_keys.discard(cache_key)

                # If there are no [other] registered cache keys associated with
                # this load key, then we won't have to load any data for that
                # key (i.e., we can drop it).
                if not pending_cache_keys:
                    self._pending_load_keys.discard(load_key)

                # Allow :py:meth:`__setitem__` to override values more than
                # once.
                self._override_cache_keys.discard(cache_key)

                self._replace_cached_value(cache_key, cache_value)
                self._override_cache_keys.add(cache_key)

    def _replace_cached_value(self,
            cache_key: typing.Hashable,
            cache_value: typing.Any,
    ) -> None:
        """
        Replaces a value in the cache.

        You may override this method in a subclass if desired.

        Note: this method must *not* raise an exception.
        """
        self._add_to_cache(cache_key, cache_value)

    def _add_to_cache(self,
            cache_key: typing.Hashable,
            cache_value: typing.Any,
    ) -> None:
        """
        Adds a value to the cache, if it has not been overridden by
        :py:meth:`__setitem__`.
        """
        if cache_key not in self._override_cache_keys:
            super(MutableDataProviderMixin, self) \
                ._add_to_cache(cache_key, cache_value)



from unittest import TestCase

from data_providers import BaseDataProvider, MutableDataProviderMixin


class DataProviderTestCase(TestCase):
    """
    Tests various aspects of data provider implementations (effectively acts as
    a test case for :py:class:`BaseDataProvider`).
    """

    def test_computed_keys(self):
        """
        The data provider uses different keys for loading and caching.
        """

        class TestDataProvider(BaseDataProvider):
            def gen_load_key(self, value):
                # Use profession to load data from the backend.
                return value['profession']

            def gen_cache_key(self, value):
                # Cache results by last name.
                return value['lastName']

            def fetch_from_backend(self, load_keys):
                # This roughly simulates a scenario where we have to hit a
                # different backend depending on the load key.
                # For example, the load key could be the name of the R function
                # to invoke, the name of the database or collection to execute
                # the query against, etc.
                for lk in load_keys:
                    method = getattr(self, 'get_{job}s'.format(job=lk), None)
                    if method:
                        for key, data in method().items():
                            yield key, data

            def gen_empty_result(self):
                return {'firstName': '*unknown*'}

            @staticmethod
            def get_historians():
                return {
                    'Brody': {'firstName': 'Marcus'},
                    'Jones': {'firstName': 'Henry'},
                }

            @staticmethod
            def get_nazi_stooges():
                return {
                    'Belloq': {'firstName': 'René'},
                    'Donovan': {'firstName': 'Walter'},
                }

        users = [
            {'lastName': 'Jones', 'profession': 'historian'},
            {'lastName': 'Brody', 'profession': 'historian'},
            {'lastName': 'Belloq', 'profession': 'nazi_stooge'},
            {'lastName': 'Donovan', 'profession': 'nazi_stooge'},

            # The backend won't return any data for this user.
            {'lastName': 'Ravenwood', 'profession': 'adventurer'},
        ]

        data_provider = TestDataProvider()
        data_provider.register(users)

        for user_data in users:
            user_data.update(data_provider[user_data])

        self.assertDictEqual(
            users[0],

            {
                'firstName': 'Henry',
                'lastName': 'Jones',
                'profession': 'historian',
            },
        )

        self.assertDictEqual(
            users[1],

            {
                'firstName': 'Marcus',
                'lastName': 'Brody',
                'profession': 'historian',
            },
        )

        self.assertDictEqual(
            users[2],

            {
                'firstName': 'René',
                'lastName': 'Belloq',
                'profession': 'nazi_stooge',
            },
        )

        self.assertDictEqual(
            users[3],

            {
                'firstName': 'Walter',
                'lastName': 'Donovan',
                'profession': 'nazi_stooge',
            },
        )

        self.assertDictEqual(
            users[4],

            {
                # See :py:meth:`TestDataProvider.gen_empty_result` above.
                'firstName': '*unknown*',

                'lastName': 'Ravenwood',
                'profession': 'adventurer',
            },
        )

    def test_load_key_null(self):
        """
        A value has a null load key.
        """

        class TestDataProvider(BaseDataProvider):
            def gen_load_key(self, value):
                if value == 'bravo':
                    return None
                return value

            def fetch_from_backend(self, load_keys):
                return {
                    # Include data for ``bravo`` in the result, just to make
                    # sure that we aren't cheating.
                    'alpha': {'firstName': 'Henry'},
                    'bravo': {'firstName': 'Marcus'},
                }

            def gen_empty_result(self):
                return {}

        users = ['alpha', 'bravo']

        data_provider = TestDataProvider()
        data_provider.register(users)

        self.assertDictEqual(data_provider['alpha'], {'firstName': 'Henry'})

        # The load key for ``bravo`` is null, so the data provider does not
        # return any data for this key, even though technically the backend
        # did.
        self.assertDictEqual(data_provider['bravo'], {})

    def test_cache_key_null(self):
        """
        A value has a null cache key.
        """

        class TestDataProvider(BaseDataProvider):
            def gen_load_key(self, value):
                return value['lastName']

            def gen_cache_key(self, value):
                if value['profession'] == 'nazi_stooge':
                    return None

                return value['lastName']

            def gen_empty_result(self):
                return {}

            def fetch_from_backend(self, load_keys):
                return {
                    # Include data for Nazi stooges in the result, just to make
                    # sure that we aren't cheating.
                    'Jones': {'firstName': 'Henry'},
                    'Belloq': {'firstName': 'René'},
                }

        users = [
            # Control Group
            {'lastName': 'Jones', 'profession': 'historian'},
            {'lastName': 'Belloq', 'profession': 'archaeologist'},

            # Experimental Group
            {'lastName': 'Belloq', 'profession': 'nazi_stooge'},
        ]

        data_provider = TestDataProvider()
        data_provider.register(users)

        # The first two users generate a non-null cache key, so the data
        # provider returns data for them.
        self.assertDictEqual(data_provider[users[0]], {'firstName': 'Henry'})
        self.assertDictEqual(data_provider[users[1]], {'firstName': 'René'})

        # The third user has a non-null load key, so the data provider was able
        # to *load* data for it.  But, it has a null cache key, so the data
        # provider can't *return* any data for it.
        self.assertDictEqual(data_provider[users[2]], {})


class MutableDataProviderTestCase(TestCase):
    """
    Tests various aspects of mutable data provider implementation (effectively
    acts as a test case for :py:class:`MutableDataProviderMixin`).
    """

    def test_add_arbitrary_value(self):
        """
        Adding an arbitrary value to the data provider's cache.
        """

        class TestDataProvider(MutableDataProviderMixin, BaseDataProvider):
            def fetch_from_backend(self, load_keys):
                return {
                    'alpha': {'firstName': 'Henry'},
                    'bravo': {'firstName': 'Marcus'},
                }

        users = ['alpha', 'bravo']

        data_provider = TestDataProvider()
        data_provider.register(users)

        # If we try to get data for an unregistered value, we get an exception
        # as normal.
        with self.assertRaises(ValueError):
            # noinspection PyStatementEffect
            data_provider['charlie']

        # However, if we specify a cache value explicitly, then we can retrieve
        # it - even if we haven't loaded anything yet.
        data_provider['charlie'] = {'firstName': 'Sallah'}

        self.assertDictEqual(
            data_provider['charlie'],
            {'firstName': 'Sallah'},
        )

        # We can override this value any time we want to.
        data_provider['charlie'] = {'firstName': 'Marion'}

        self.assertDictEqual(
            data_provider['charlie'],
            {'firstName': 'Marion'},
        )

        # Other values continue to work as expected.
        self.assertDictEqual(
            data_provider['alpha'],
            {'firstName': 'Henry'},
        )

        with self.assertRaises(ValueError):
            # noinspection PyStatementEffect
            data_provider['delta']

    def test_override_backend_value(self):
        """
        Overriding data returned by the backend.
        """

        class TestDataProvider(MutableDataProviderMixin, BaseDataProvider):
            def fetch_from_backend(self, load_keys):
                return {
                    'alpha': {'firstName': 'Henry'},
                    'bravo': {'firstName': 'Marcus'},
                }

        users = ['alpha', 'bravo']

        data_provider = TestDataProvider()
        data_provider.register(users)

        # Specify an explicit cache value for ``bravo``.
        data_provider['bravo'] = {'firstName': 'Sallah'}

        # Explicitly invoke the data provider's backend.
        data_provider.warm_cache()

        # The backend cannot replace an overridden value.
        self.assertDictEqual(
            data_provider['bravo'],
            {'firstName': 'Sallah'},
        )

        # But we can override it any time we want.
        data_provider['bravo'] = {'firstName': 'Marion'}

        self.assertDictEqual(
            data_provider['bravo'],
            {'firstName': 'Marion'},
        )

        # Other values continue to work as expected.
        self.assertDictEqual(
            data_provider['alpha'],
            {'firstName': 'Henry'},
        )

        with self.assertRaises(ValueError):
            # noinspection PyStatementEffect
            data_provider['delta']
